@@ -2,7 +2,6 @@
 #include <Python.h>
 
 #include <cassert>
-#include <limits.h>
 
 #ifdef __linux__
 #    include <link.h>
@@ -16,7 +15,6 @@
 #include <mutex>
 #include <type_traits>
 #include <unistd.h>
-#include <utility>
 
 #include "compat.h"
 #include "exceptions.h"
@@ -57,11 +55,20 @@ namespace memray::tracking_api {
 
 MEMRAY_FAST_TLS thread_local bool RecursionGuard::isActive = false;
 
+static inline thread_id_t
+generate_next_tid()
+{
+    static std::atomic<thread_id_t> s_tid_counter = 0;
+    return ++s_tid_counter;
+}
+
+MEMRAY_FAST_TLS thread_local thread_id_t t_tid = generate_next_tid();
+
 static inline thread_id_t
 thread_id()
 {
-    return reinterpret_cast<thread_id_t>(pthread_self());
-};
+    return t_tid;
+}
 
 // Tracker interface
 
@@ -77,9 +84,7 @@ thread_id()
 class PythonStackTracker
 {
   private:
-    PythonStackTracker()
-    {
-    }
+    PythonStackTracker() = default;
 
     enum class FrameState {
         NOT_EMITTED = 0,
@@ -107,7 +112,7 @@ class PythonStackTracker
     void emitPendingPushesAndPops();
     void invalidateMostRecentFrameLineNumber();
     int pushPythonFrame(PyFrameObject* frame);
-    void popPythonFrame(PyFrameObject* frame);
+    void popPythonFrame();
 
     void installGreenletTraceFunctionIfNeeded();
     void handleGreenletSwitch(PyObject* from, PyObject* to);
@@ -213,8 +218,9 @@ PythonStackTracker::invalidateMostRecentFrameLineNumber()
     // As bytecode instructions are executed, the line number in the most
     // recent Python frame can change without us finding out. Cache its line
     // number, but verify it the next time this frame might need to be emitted.
-    if (!d_stack->empty()
-        && d_stack->back().state == FrameState::EMITTED_AND_LINE_NUMBER_HAS_NOT_CHANGED) {
+    if (d_stack && !d_stack->empty()
+        && d_stack->back().state == FrameState::EMITTED_AND_LINE_NUMBER_HAS_NOT_CHANGED)
+    {
         d_stack->back().state = FrameState::EMITTED_BUT_LINE_NUMBER_MAY_HAVE_CHANGED;
     }
 }
@@ -291,15 +297,11 @@ PythonStackTracker::pushLazilyEmittedFrame(const LazilyEmittedFrame& frame)
 }
 
 void
-PythonStackTracker::popPythonFrame(PyFrameObject* frame)
+PythonStackTracker::popPythonFrame()
 {
     installGreenletTraceFunctionIfNeeded();
 
-    // Note: We check if frame == d_stack->back().frame because Cython could
-    // have called our tracing function with profiled Cython calls that we
-    // later discarded in favor of the interpreter's stack when a new tracker
-    // was installed. If so, we need to ignore the Cython frame pops.
-    if (!d_stack || d_stack->empty() || frame != d_stack->back().frame) {
+    if (!d_stack || d_stack->empty()) {
         return;
     }
 
@@ -363,9 +365,9 @@ PythonStackTracker::installGreenletTraceFunctionIfNeeded()
     if (!warned) {
         warned = true;
 
-        PyObject* ret = PyObject_CallMethod(_memray, "print_greenlet_warning", nullptr);
-        Py_XDECREF(ret);
-        if (!ret) {
+        PyObject* res = PyObject_CallMethod(_memray, "print_greenlet_warning", nullptr);
+        Py_XDECREF(res);
+        if (!res) {
             PyErr_Print();
             _exit(1);
         }
@@ -373,7 +375,7 @@ PythonStackTracker::installGreenletTraceFunctionIfNeeded()
 }
 
 void
-PythonStackTracker::handleGreenletSwitch(PyObject*, PyObject*)
+PythonStackTracker::handleGreenletSwitch(PyObject* from, PyObject* to)
 {
     RecursionGuard guard;
 
@@ -386,6 +388,25 @@ PythonStackTracker::handleGreenletSwitch(PyObject*, PyObject*)
         emitPendingPushesAndPops();
     }
 
+    // Save current TID on old greenlet. Print errors but otherwise ignore them.
+    PyObject* tid = PyLong_FromUnsignedLong(t_tid);
+    if (!tid || 0 != PyObject_SetAttrString(from, "_memray_tid", tid)) {
+        PyErr_Print();
+    }
+    Py_XDECREF(tid);
+
+    // Restore TID from new greenlet, or generate a new one. Ignore errors:
+    // maybe we haven't seen this TID before, or maybe someone overwrote our
+    // attribute, but either way we can recover by generating a new one.
+    tid = PyObject_GetAttrString(to, "_memray_tid");
+    if (!tid || !PyLong_CheckExact(tid)) {
+        PyErr_Clear();
+        t_tid = generate_next_tid();
+    } else {
+        t_tid = PyLong_AsUnsignedLong(tid);
+    }
+    Py_XDECREF(tid);
+
     // Re-create our TLS stack from our Python frames, most recent last.
     // Note: `frame` may be null; the new greenlet may not have a Python stack.
     PyFrameObject* frame = PyEval_GetFrame();
@@ -423,7 +444,10 @@ PythonStackTracker::pythonFrameToStack(PyFrameObject* current_frame)
             return {};
         }
 
-        stack.push_back({current_frame, {function, filename, 0}, FrameState::NOT_EMITTED});
+        // If native tracking is not enabled, treat every frame as an entry frame.
+        // It doesn't matter to the reader, and is more efficient.
+        bool entry = !s_native_tracking_enabled || compat::isEntryFrame(current_frame);
+        stack.push_back({current_frame, {function, filename, 0, entry}, FrameState::NOT_EMITTED});
         current_frame = compat::frameGetBack(current_frame);
     }
 
@@ -453,12 +477,6 @@ PythonStackTracker::recordAllStacks()
         }
     }
 
-    // Throw away all but the most recent frame for this thread.
-    // We ignore every stack frame above `Tracker.__enter__`.
-    PyThreadState* tstate = PyThreadState_Get();
-    assert(stack_by_thread[tstate].size() >= 1);
-    stack_by_thread[tstate].resize(1);
-
     std::unique_lock<std::mutex> lock(s_mutex);
     s_initial_stack_by_thread.swap(stack_by_thread);
 
@@ -540,6 +558,7 @@ Tracker::Tracker(
         pthread_atfork(&prepareFork, &parentFork, &childFork);
     });
 
+    d_writer->setMainTidAndSkippedFrames(thread_id(), computeMainTidSkip());
     if (!d_writer->writeHeader(false)) {
         throw IoError{"Failed to write output header"};
     }
@@ -566,10 +585,17 @@ Tracker::~Tracker()
     PythonStackTracker::s_native_tracking_enabled = false;
     d_background_thread->stop();
     d_patcher.restore_symbols();
-    if (d_trace_python_allocators) {
-        unregisterPymallocHooks();
+    if (Py_IsInitialized() && !_Py_IsFinalizing()) {
+        PyGILState_STATE gstate;
+        gstate = PyGILState_Ensure();
+
+        if (d_trace_python_allocators) {
+            unregisterPymallocHooks();
+        }
+        PythonStackTracker::removeProfileHooks();
+
+        PyGILState_Release(gstate);
     }
-    PythonStackTracker::removeProfileHooks();
     d_writer->writeTrailer();
     d_writer->writeHeader(true);
     d_writer.reset();
@@ -706,7 +732,7 @@ Tracker::childFork()
 
     // If we inherited an active tracker, try to clone its record writer.
     std::unique_ptr<RecordWriter> new_writer;
-    if (old_tracker && old_tracker->isActive() && old_tracker->d_follow_fork) {
+    if (old_tracker && memray::tracking_api::Tracker::isActive() && old_tracker->d_follow_fork) {
         new_writer = old_tracker->d_writer->cloneInChildProcess();
     }
 
@@ -734,6 +760,26 @@ Tracker::childFork()
     RecursionGuard::isActive = false;
 }
 
+size_t
+Tracker::computeMainTidSkip()
+{
+    // Determine how many frames from the current stack to elide from our
+    // reported stack traces. This avoids showing the user frames above the one
+    // that called `Tracker.__enter__`.
+    assert(PyGILState_Check());
+
+    PyFrameObject* frame = PyEval_GetFrame();
+
+    size_t num_frames = 0;
+    while (frame) {
+        ++num_frames;
+        frame = compat::frameGetBack(frame);
+    }
+
+    assert(num_frames > 0);
+    return num_frames - 1;
+}
+
 void
 Tracker::trackAllocationImpl(void* ptr, size_t size, hooks::Allocator func)
 {
@@ -851,13 +897,13 @@ Tracker::updateModuleCacheImpl()
     uint32_t c = _dyld_image_count();
     for (uint32_t i = 0; i < c; i++) {
         const struct mach_header* header = _dyld_get_image_header(i);
-        uintptr_t slide = static_cast<uintptr_t>(_dyld_get_image_vmaddr_slide(i));
+        auto slide = static_cast<uintptr_t>(_dyld_get_image_vmaddr_slide(i));
         const char* image_name = _dyld_get_image_name(i);
         std::vector<Segment> segments;
 
-        const segment_command_t* current_segment_cmd = nullptr;
+        const segment_command_t* current_segment_cmd;
         uintptr_t current_cmd = reinterpret_cast<uintptr_t>(header) + sizeof(mach_header_t);
-        for (uint i = 0; i < header->ncmds; i++, current_cmd += current_segment_cmd->cmdsize) {
+        for (uint j = 0; j < header->ncmds; j++, current_cmd += current_segment_cmd->cmdsize) {
             current_segment_cmd = reinterpret_cast<const segment_command_t*>(current_cmd);
             if (current_segment_cmd->cmd == ARCH_LC_SEGMENT) {
                 segments.emplace_back(Segment{current_segment_cmd->vmaddr, current_segment_cmd->vmsize});
@@ -1067,12 +1113,16 @@ PyTraceFunction(
         return 0;
     }
 
+    if (frame != PyEval_GetFrame()) {
+        return 0;
+    }
+
     switch (what) {
         case PyTrace_CALL: {
             return PythonStackTracker::get().pushPythonFrame(frame);
         }
         case PyTrace_RETURN: {
-            PythonStackTracker::get().popPythonFrame(frame);
+            PythonStackTracker::get().popPythonFrame();
             break;
         }
         default:
